Program Main

    !Program: MRFIT (Mie Resonance Fitting)
    !Author: Thomas C. Preston 
    !Contact: thomas.preston@mcgill.ca

 	!Version 1.0: October.08.2015 
	!First public release. 
 
    !Description:
    !Finds the size and RI of a sphere by fitting Mie resonances. The RI is defined by: 
    !m = m0 + m1*(v-v0) + m2*(v-v0)^2 + ... + mk*(v-v0)^k,
    !where v is the wavenumber and v0 is some point on the spectrum (usually the center)
    !Alternatively the RI maybe defined using a Cauchy type equation
    !m = m0 + m1*v^2 + m2*v^4 + ... + mk*v^(2k) 
    
    !For more details see: 
    !Preston and Reid (2015) JOSA A. 32: 2210–2217.
    
    !All code written by TCP is covered by the GNU General Public License. 
    !The only code not written by TCP is the complex Bessel function module, 
    !complexbessel.f90, see module header for source.
    
    use working_precision
    use readandwrite
    use resonances
    use analysis
    use reduced

    Implicit None
  
    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file with index and radius only
    
    real(wp), allocatable :: medRI(:) !The RI of the medium
    real(wp), allocatable :: obsModes(:) !A list that contains ALL the observed modes (in wavenumber)
    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    real(wp), allocatable :: modeTable(:,:,:,:,:) !table that stores x, q, b for quick access
                            !modeTable(m,N,L,P,x)
    real(wp), allocatable :: p(:) !contains the solved parameters (radius, m0, m1, ...)
    integer, allocatable :: nModes(:,:) !the number of modes per frame and the starting posn in the obsModes list
    integer, allocatable :: minN(:) !The minimum mode number for a given mode order
    integer, allocatable :: maxN(:) !The maximum mode number for a given mode order 
    integer, allocatable :: ModeSet(:,:) !the best mode set
    integer, allocatable :: Known(:,:) !If the mode assignment is known it can be used as an input parameter, this stores it.
    
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    
    real(wp) :: minRad, maxRad, iRad !Range of possible radii
    real(wp) :: minRI, maxRI, iRI !Range of possible refractive index
    real(wp) :: minDis, maxDis, iDis !Range of possible refractive index
    real(wp) :: v0 !The midpoint of the spectrum -- wavenumber
    real(wp) :: cRI !the current RI
    real(wp) :: dm0,dm1 !the values of m0 and m1 from the initial linear dispersion fit     
    real(wp) :: t1, t2 !used to track time
   
    integer :: cModes, sPosn !Number of modes in current frame and the starting position in obsModes
    integer :: sframe, fframe, iframe, cframe !starting frame, ending frame, interval for frames
    integer :: minL, maxL !starting and final mode order
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + mk*(v-v0)^k)
    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)
    integer :: nMax, nFrames    !the total number of modes listed in the input file and the total number of frames
    integer :: nCalcModes !the number of calculated modes for the current frame
    integer :: fRed !should the reduced search be used  (0:No, 1:Yes)
    integer :: fCau !Only use even terms in RI expansion m0 + m2*(v-v0)^2 + m4*(v-v0)^4 + ... (0:No, 1:Yes)
    integer :: fKnown !The mode assignments are known/listed in the input file (0:No, 1:Yes)
    real(wp) :: RIfact, radfact !factors used in the RI and radius for the reduced search
    real(wp) :: origVal(6) !stores the original values of minRad,maxRad,iRad,minRI,maxRI,iRI during reduced search
    integer :: mlow !associated with the reduced search, because modes are read from a table when the minRI changes low index of table changes
    
    integer :: nRI, nRad, nDis !number of steps in RI, radius, and dispersion
    
    integer :: jsw !see readandwrite for details
	real(wp) :: jswMin, jswMax !The maximum and minimum wavenumber for the jsw mode
    
    integer :: i,j,N,L,T,m

        
    !Fill the Airy array with the roots to the Airy function
    Call AiryRoots(Airy)    
  
    Call ReadInput(peaklistname,outputname,altoutputname,sframe,fframe,iframe, &
                   minRad,maxRad,iRad,minRI,maxRI,iRI,minDis,maxDis,iDis,v0,minL,maxL,medRI,k,ks, &
                   fRed,RIfact,radfact,fCau,jsw,jswMin,jswMax,fKnown)
  
    !Import frame list and format
    Call FormatInput(peaklistname,obsModes,nModes,nFrames,nMax,fKnown,Known,ks,medRI,minRI,maxRI,iRI)
  
    !Number of frames
    If (fframe.NE.-1) Then
        nframes = fframe
    End if
    
    allocate(p(k+2))
    allocate(minN(minL:maxL)) 
    allocate(maxN(minL:maxL))

    !Find the minimum and maximum wavenumber in the entire set
    allocate(vexp(2))
    vexp(1)=minval(obsModes)
    vexp(2)=maxval(obsModes)
    If(jsw.EQ.1 .OR. jsw.EQ.2) then
        vexp(1)=jswMin
        vexp(2)=jswMax
    End if
    
    Call Bounds(minRad,maxRad,minRI,maxRI,minL,maxL,minN,maxN,vexp,Airy)
    deallocate(vexp)

    !Write the headers to the output file
    Call WriteHeaders(outputname,altoutputname,k,minL,maxL,minN,maxN,medRI,ks)

    !number of points during search    
    nRad = NInt((maxRad-minRad)/iRad) + 1
    nRI = NInt((maxRI-minRI)/iRI) + 1
    nDis = NInt((maxDis-minDis)/iDis) + 1

    !Build table with modes at the midpoint of the RI range
    allocate(modeTable(0:nRI+1,minval(minN):maxval(maxN),minL:maxL,0:1,3))
    modeTable=dcmplx(0.0_wp,0.0_wp)

    !For the modes of interest find not only x, but also q and b in x=RI*q+b

    If (fRed.EQ.1) Then
        Write(*,'(a)') "Using reduced search."
    End if 

    If (jsw.EQ.0 .OR. jsw.EQ.1) then
		Write(*,'(a)') "Constructing table of modes."
	
		Do m = 0, nRI+1
			cRI = real(m-1)*iRI + minRI
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						modeTable(m,N,L,T,1)=MieRes(N,Airy(L),cRI,T) !x
					End do
				End do
			End do
		End do
		
		Do m = 1, nRI
			cRI = real(m-1)*iRI + minRI
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						modeTable(m,N,L,T,2)=(modeTable(m+1,N,L,T,1)-modeTable(m-1,N,L,T,1))/(2*iRI) !q
						modeTable(m,N,L,T,3)=modeTable(m,N,L,T,1)-modeTable(m,N,L,T,2)*cRI !b
					End do
				End do
			End do
		End do
		
		Write(*,'(a)') "Table of modes is built."
		
		If (jsw.EQ.1) Then !write the mode table to an output file to be read later
			Call jswReadWrite(jsw,modeTable,minN,maxN,minL,maxL,nRI)
		End if
		
	Else if (jsw.EQ.2) Then !read an already constructed mode table
		Call jswReadWrite(jsw,modeTable,minN,maxN,minL,maxL,nRI)
	End if

    Do cFrame = sframe, nframes, iframe
    
        cModes = nModes(1,cframe) 
        !You need at least k + 2 modes in a frame to fit
        If (cModes.GE.(k+2)) Then
            
            Call cpu_time(t1)
            
            sPosn = nModes(2,cFrame)
            
            allocate(vexp(cModes))
            allocate(ModeSet(cModes,3))

            !read the modes of the current frame into the array
            Do i = 1,cModes
            	j = i + sPosn - 1
                vexp(i) = obsModes(j)
            End do
            
            !set the solution vector to 0
            p=0.0_wp
            
            If (fKnown.EQ.0) Then				
				mlow=1
				If (fRed.EQ.1) Then
					!do the reduced search
					!note the dispersion has been set to 0 with 1 step for the first call
					Call AssignModes(cModes,minL,maxL,minN,maxN,vexp,v0,modeTable,p,&
									 nRad,iRad,minRad,nRI,iRI,minRI,1,iDis,0.0_wp,ModeSet,mlow)           
					Call storeParameters(minRad,maxRad,iRad,nRad,minRI,maxRI,iRI,nRI,origVal,0)
					Call reducedRadRI(p,minRad,maxRad,iRad,nRad,minRI,maxRI,iRI,nRI,origVal,RIfact,radfact,mlow)
				End if
	
				Call AssignModes(cModes,minL,maxL,minN,maxN,vexp,v0,modeTable,p,&
								  nRad,iRad,minRad,nRI,iRI,minRI,nDis,iDis,minDis,ModeSet,mlow)
	
				If (fRed.EQ.1) Then
					!reset the minRad and maxRad
					Call storeParameters(minRad,maxRad,iRad,nRad,minRI,maxRI,iRI,nRI,origVal,1)
				End if

			Else if (fKnown.EQ.1) Then		
				!If the mode assignments are known use them
				Call AssignKnown(cModes,sPosn,ModeSet,Known,p,minRad,maxRad,minRI,maxRI)			
			End if

            Call IdentifyParameters(vexp,v0,cModes,minRI,maxRI,iRI,nRI,modeTable,ModeSet,k,p,Airy,fCau)
            
            Call WriteOutput(outputname,altoutputname,cFrame,cModes,p,k,ModeSet,vexp,Airy,v0,fCau,medRI,ks)
            
            Call cpu_time(t2)
            
            Call WriteStatus(cFrame,nFrames,t2-t1)
                        
            deallocate(ModeSet)
            deallocate(vexp)
            
        Else
        
            Call WriteStatus(cFrame,nFrames,-1.0_wp)
        
        End if
    
    End do   

    deallocate(minN)
    deallocate(maxN)
    deallocate(modeTable)
    deallocate(p)
    deallocate(nModes)
    If (fKnown.EQ.1) Then
		deallocate(Known)
	End if
    
 
End Program Main